#include <stdio.h>
#include <signal.h>
#define TRUE 1
#define FALSE 0
#define boolean int

void break_handler(int signum)
{
    signal(SIGINT,break_handler);
    put_rs232(3);
}

void string_toupper(char *str)
{
    int j;
    for (j=0;str[j];j++) str[j]=toupper(str[j]);
}

/***********************************************************/
#define SIZE 0x2000
char buffer[SIZE];
volatile int buffer_put_ptr, buffer_get_ptr;

#define CHAR_RECEIVED (buffer_put_ptr!=buffer_get_ptr)

char get_rs232()
{
    char c=buffer[buffer_get_ptr++];
    buffer_get_ptr&=SIZE-1;
    return c;
}

extern void put_rs232(char c);
/***********************************************************/

int line,column;
int width=80, height=25;
boolean accent_set=FALSE;
boolean no_snow=FALSE;
boolean no_clock=FALSE;
boolean cursor_on=TRUE;
extern void scroll(),clear_screen();
extern show_cursor(),hide_cursor();
extern void put_screen(int offset,int value);
extern int get_screen(int offset);
extern void set_cursor_address(int offset);

disable_clock()
{
    outp(0x21,inp(0x21)|1);
}

enable_clock()
{
    outp(0x21,inp(0x21)&0xFE);
}

void display_ascii(char c)
{
    if (c==96) c='c'; /* (C) */
    else if (c==95) c='';
    else if (c==126) c='';
    if (accent_set)
	switch(c) {
	    case  64: c=133; break; /*  */
	    case  92: c=135; break; /*  */
	    case 123: c=130; break; /*  */
	    case 124: c=151; break; /*  */
	    case 125: c=138; break; /*  */
	    case 126: c=136; break; /*  */
	}
    put_screen((line*80+column)*2,c);
}

void clear_line()
{
    int i, offset=(line*80+column)*2;
    for (i=0;i<width*2;i+=2) {
	put_screen(offset+i,' ');
	put_screen(offset+i+1,0x0F);
    }
}

void move_cursor_right()
{
    column++;
    if (column==width) {
	column=0;
	line++;
	if (line==height) {
	    line--;
	    scroll();
	}
    }
}

void set_inversed_video()
{
    int offset=(line*80+column)*2+1;
    put_screen(offset,get_screen(offset)^0x7F);
}

void set_attribute(int attr)
{
    static int foreground[8]={0,4,2,14,1,13,11,15};
    static int background[8]={0,4,2,6,1,5,3,7};
    int categ=attr/8, color=attr%8;
    int i,offset=line*80*2;
    switch (categ) {
	case 0:
	    for (i=column;i<width;i++) {
		put_screen(offset+2*i+1,
			get_screen(offset+2*i+1)&0xF0|foreground[color]);
	    }
	    break;
	case 1:
	    if (attr&4)
		for (i=column;i<width;i++) {
		    put_screen(offset+2*i+1,get_screen(offset+2*i+1)|0x80);
		}
	    break;
	case 2:
	    for (i=column;i<width;i++) {
		put_screen(offset+2*i+1,
			get_screen(offset+2*i+1)&0x0F|background[color]<<4);
	    }
	    break;
    }
}

void display(unsigned char c)
{
    static boolean escape=FALSE, us_sequence=FALSE, us_sequence2=FALSE;
    static inverse=FALSE;
    if (escape) {
	set_attribute(c);
	display_ascii(' ');
	move_cursor_right();
	escape=FALSE;
    } else if (us_sequence) {
	if (us_sequence2) {
	    column=c-64;
	    us_sequence=us_sequence2=FALSE;
	} else {
	    int i;
	    line=c-64;
	    if (line>=height) {
		for(i=0;i<=line-height;i++) scroll();
		line=height-1;
	    }
	    us_sequence2=TRUE;
	}
    } else {
	if (c<32 || c==127) {
	  switch (c) {
	    case 7: putch(7); break;
	    case 127:
	    case 8:
		column--;
		if (column>=0) break;
		column=width-1;
	    case 11:
		line--;
		if (line<0) {
		    column=0;
		    line=0;
		}
		break;
	    case 9:
		column++;
		if (column<width) break;
		column=0;
	    case 10:
		line++;
		if (line==height) {
		    line--;
		    scroll();
		}
		break;
	    case 12: clear_screen();
	    case 30: line=column=0; break;
	    case 13: column=0; break;
	    case 14: clear_line(); break;
	    case 22: inverse=~inverse; break;
	    case 26:
	    case 28:
	    case 27: escape=TRUE; break;
	    case 31: us_sequence=TRUE; break;
	  }
	  if (c==127) display_ascii(' ');
	} else if (c>=0xA0) {
	    set_inversed_video();
	    display_ascii(c-128);
	    move_cursor_right();
	} else if (c>=0x80) {
	    set_attribute(c-128);
	    display_ascii(' ');
	    move_cursor_right();
	} else if (c>=32 && c<127) {
	    if (inverse) set_inversed_video();
	    display_ascii(c);
	    move_cursor_right();
	}
    }
    if (!us_sequence2) set_cursor_address((line*80+column)*2);
}

char usertab[78][18];

void funct_sequence(int n)
{
	char *str=usertab[n];
	while (*str) { 
		put_rs232(*str++);
		while (!CHAR_RECEIVED)
			; /* waits remote echo of character */
		display(get_rs232());
	}
}

void terminal()
{
    char c;
    for(;;) {
	if (CHAR_RECEIVED) display(get_rs232());
	if (kbhit()) {
	    c=getch();
	    switch (c) {
		case 0:
		    c=getch();
		    if (c>=59 && c<=67) funct_sequence(c-59); /* F1 to F9 */
		    else if (c>=84 && c<113) funct_sequence(10+c-84); /* shift F1 to F10, ctrl..., alt.. */
		    else if (c>=120 && c<=131) funct_sequence(40+c-120); /* 1st line */
		    else if (c>=16 && c<=25) funct_sequence(52+c-16); /* 2nd line */
		    else if (c>=30 && c<=38) funct_sequence(62+c-30); /* 3rd line */
		    else if (c>=44 && c<=50) funct_sequence(71+c-44); /* 4th line */
		    else switch (c) {
			case 0x44: return; /* F10 */
			case 0x48: put_rs232(0x0B); break; /* UP */
			case 0x4B: put_rs232(0x08); break; /* LEFT */
			case 0x4D: put_rs232(0x09); break; /* RIGHT */
			case 0x50: put_rs232(0x0A); break; /* DOWN */
		    }
		    break;
		case 1: put_rs232(get_screen((line*80+column)*2)); break;
		case 8: put_rs232(127); break; /* unfortunately, backspace is Ctrl-H on the PC */
		case 17: cursor_on^=1; 
			if (cursor_on) { 
				set_cursor_address((line*80+column)*2);
				show_cursor(); 
			} else {
				hide_cursor();
				set_cursor_address(-1);
			}
			break;
		case 130: put_rs232(123); break; /*  */
		case 133: put_rs232( 64); break; /*  */
		case 135: put_rs232( 92); break; /*  */
		case 136: put_rs232(126); break; /*  */
		case 138: put_rs232(125); break; /*  */
		case 151: put_rs232(124); break; /*  */
		case 177: put_rs232(126); break; /*  */
		default: if (c<128) put_rs232(c);
	    }
	}
    }
}

/***********************************************************/

int speed=9600;
int data_bits=8;
char parity='N';
int stop_bits=1;
int rs232_base_addr=0x3F8;
int interrupt_number=4;

parse_args(int argc, char *argv[])
{
    int i,j;
    for (i=1; i<argc; i++) {
	string_toupper(argv[i]);
	if (strncmp(argv[i],"-COM",4)==0)
		switch (argv[i][4]) {
		    case '1':
			rs232_base_addr=0x3F8;
			interrupt_number=4;
			break;
		    case '2':
			rs232_base_addr=0x2F8;
			interrupt_number=3;
			break;
		    case '3':
			rs232_base_addr=0x3E8;
			interrupt_number=4;
			break;
		    case '4':
			rs232_base_addr=0x2E8;
			interrupt_number=3;
			break;
		    default:
			printf("I don't know this COM port...\n");
			exit(1);
		}
	else if (strncmp(argv[i],"-INT=",5)==0)
		sscanf(&argv[i][5],"%d",&interrupt_number);
	else if (strncmp(argv[i],"-BAUD=",6)==0)
		sscanf(&argv[i][6],"%d",&speed);
	else if (strncmp(argv[i],"-DATA=",6)==0)
		sscanf(&argv[i][6],"%d",&data_bits);
	else if (strncmp(argv[i],"-STOP=",6)==0)
		sscanf(&argv[i][6],"%d",&stop_bits);
	else if (strncmp(argv[i],"-PAR=",5)==0)
		sscanf(&argv[i][5],"%c",&parity);
	else if (strcmp(argv[i],"-ACC")==0) accent_set=TRUE;
	else if (strcmp(argv[i],"-SNOW")==0) no_snow=TRUE;
	else if (strcmp(argv[i],"-CLOCK")==0) no_clock=TRUE;
	else {
	    printf("Usage: %s [options]\n",argv[0]);
	    printf("\n");
	    printf("Options:\n");
	    printf("\t-COMn selects a COM port with default settings (default: COM1)\n");
	    printf("\t-INT=n changes the interrupt level for the selected COM port\n");
	    printf("\t-BAUD=n selects speed (default: 9600)\n");
	    printf("\t-DATA=n selects data bits (default: 8 bits)\n");
	    printf("\t-PAR={N,E,O} selects parity (default: No parity)\n");
	    printf("\t-STOP={1,2} selects stop bits (default: 1 stop bit)\n");
	    printf("\t-ACC selects French accent set\n");
	    printf("\t-SNOW avoids snow on CGA cards\n");
	    printf("\t-CLOCK disables clock interrupt\n");
	    printf("\n");
	    printf("Examples:\n");
	    printf("\t%s -COM2\n",argv[0]);
	    printf("\t%s -COM3 -INT:9\n",argv[0]);
	    printf("\t%s -BAUD=1200 -DATA=7 -PAR=E\n",argv[0]);
	    printf("\n");
	    exit(1);
	}
    }
}

void read_funct_sequences()
{
    int n;
    char *str;
    FILE *f=fopen("userkeys.def","r");
    if (f==NULL) return;
    for (n=0;n<78;n++) {
	str=usertab[n];
	fgets(str,18,f);
	while (*str) {
		if (*str=='\n' || *str=='\r') *str='\0';
		if (*str=='') *str='\n';
		str++;
	}
    }
    fclose(f);
}

main(int argc, char *argv[])
{
    parse_args(argc,argv);
    read_funct_sequences();
    signal(SIGINT,break_handler);
    clear_screen();
    show_cursor();
    set_comm_params(speed,data_bits,parity,stop_bits);
    install_rs232();
    if (no_clock) disable_clock();
    terminal();
    if (no_clock) enable_clock();
    clear_screen();
    uninstall_rs232();
}
